"""Various utility functions."""

from __future__ import annotations

import sys
import functools
import inspect

def _run_before_decorator(before_func):
    """Returns a decorator that runs the specified `before_func` before the wrapped function."""
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Extract 'release' argument
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            if "release" in bound_args.arguments:
                release_value = bound_args.arguments["release"]
                before_func(release_value)

            return func(*args, **kwargs)

        return wrapper
    return decorator

def wrap_functions_with_release(module_name, before_func,func_list):
    """Dynamically wraps all functions with 'release' parameter in the given module."""
    module = sys.modules[module_name]
    
    for name in dir(module):
        attr = getattr(module, name)
        if callable(attr) and name != before_func.__name__ and name in func_list: # Wrap only user-facing functions with "release"
            sig = inspect.signature(attr)
            if "release" in sig.parameters:  
                setattr(module, name, _run_before_decorator(before_func)(attr))




"""Functions to fetch  geoparquet data from Overture Maps on AWS"""
from importlib import resources
import requests
from io import StringIO
from uuid import uuid4
from typing import Union
from geopandas import GeoDataFrame
from pandas import read_csv, DataFrame
from ._utils import wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place, schema_from_dataset
from ._category_finder import CategoryFinder

OVERTURE_CATEGORIES_URL = "https://raw.githubusercontent.com/OvertureMaps/schema/refs/heads/main/docs/schema/concepts/by-theme/places/overture_categories.csv"
OVERTURE_VERSIONS_URL = "https://github.com/tbanken/pyplaces/blob/main/src/pyplaces/releases/overture/releases.txt"

def _check_or_get_release(release: str = None, url = OVERTURE_VERSIONS_URL,latest: bool = False):
    """
    Validate if the provided Overture release is valid.
    
    Parameters
    ----------
    release : str
        The release version to validate
        
    Raises
    ------
    ValueError
        If the release version is not valid
    """
    raw_url = url.replace('github.com', 'raw.githubusercontent.com').replace('/blob/', '/')
    
    try:
        # Fetch the file content
        response = requests.get(raw_url,timeout=10)
        response.raise_for_status()
        file = StringIO(response.text)
        
        folders = [line.strip(" \n/") for line in file]
            
    except requests.RequestException as e:
        
        print(f"Error fetching file: {e}, using local version check")
        if latest:
            print("Using local version file. Current release may not be up to date.")
        with resources.files("pyplaces").joinpath("releases/overture/releases.txt").open( "r",encoding="utf-8-sig") as file:
            folders = [line.strip(" \n/") for line in file]
    if latest:
        return folders[-1]
    if release not in folders:
        raise ValueError(f"Invalid release:{release}")

OVERTURE_LATEST_RELEASE = _check_or_get_release(latest=True)

OVERTURE_MAIN_PATH = 's3://overturemaps-us-west-2/release/{release}/'
OVERTURE_BUCKET = 'overturemaps-us-west-2'
OVERTURE_REGION = 'us-west-2'

OVERTURE_PLACES_PREFIX = "theme=places/type=place/"
OVERTURE_BUILDINGS_PREFIX = "theme=buildings/type=building/"
OVERTURE_BUILDINGS_PART_PREFIX = "theme=buildings/type=building_part/"
OVERTURE_ADDRESSES_PREFIX = "theme=addresses/type=address/"
OVERTURE_TRANSPORTATION_SEGMENT_PREFIX = "theme=transportation/type=segment/"
OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX = "theme=transportation/type=connector/"
OVERTURE_BASE_PREFIX = "theme=base/type={type}"

def overture_places_from_address(address: str | tuple[float,float],
                                columns: list[str]| None = None,
                                filters: str | None = None,
                                distance: float = 500 ,
                                unit: str = "m" ,
                                release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve places data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified bounding box of the address
    """
    return from_address(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_places_from_place(address: str,columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified place
    """
    return from_place(address,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_places_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve places data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing places data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_PLACES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: str| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,building_part: bool=False) -> GeoDataFrame:
    """
    Retrieve buildings data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified bounding box of the address
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX 
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_buildings_from_place(address: str,columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified place
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_buildings_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE,building_part: bool=False)-> GeoDataFrame:
    """
    Retrieve buildings data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    building_part : bool, optional
        Whether to retrieve building parts instead of whole buildings, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing buildings data within the specified bounding box
    """
    if building_part:
        prefix = OVERTURE_BUILDINGS_PART_PREFIX
    else:
        prefix = OVERTURE_BUILDINGS_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: str| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE,connector:bool=False) -> GeoDataFrame:
    """
    Retrieve transportation data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified bounding box of the address
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_address(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_transportation_from_place(address: str,columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified place
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_place(address,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_transportation_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: str | None=None,release: str=OVERTURE_LATEST_RELEASE,connector:bool=False)-> GeoDataFrame:
    """
    Retrieve transportation data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
    connector : bool, optional
        Whether to retrieve connectors instead of segments, by default False
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing transportation data within the specified bounding box
    """
    if connector:
        prefix = OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
    else:
        prefix = OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    return from_bbox(bbox,prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    
def overture_addresses_from_address(address: str | tuple[float,float],columns: list[str]| None = None,filters: str| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve address data from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified bounding box of the address
    """
    return from_address(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)

def overture_addresses_from_place(address: str,columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified place
    """
    return from_place(address,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_addresses_from_bbox(bbox: tuple[float,float,float,float],columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve address data from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing address data within the specified bounding box
    """
    return from_bbox(bbox,OVERTURE_ADDRESSES_PREFIX,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def overture_base_from_address(address: str | tuple[float,float],base_type: str,columns: list[str]| None = None,filters: str| None = None,distance: float = 500 ,unit: str = "m" ,release: str = OVERTURE_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The address or (longitude, latitude) tuple to search for nearby places.
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water".
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, optional
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, optional
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the specified bounding box of the address.
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_address(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters,distance,unit)
    

def overture_base_from_place(address: str,base_type: str,columns: list[str]| None=None,filters: str | None =None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame:
    """
    Retrieve base data of a specific type from Overture by its address or place name.
    
    Parameters
    ----------
    address : str
        The place name to search within
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water"
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the place
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_place(address,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)
    

def overture_base_from_bbox(bbox: tuple[float,float,float,float],base_type: str,columns: list[str]| None=None,filters: str| None=None,release: str=OVERTURE_LATEST_RELEASE)-> GeoDataFrame  :
    """
    Retrieve base data of a specific type from Overture within a bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        The bounding box coordinates (min_x, min_y, max_x, max_y)
    base_type : str
        The type of base data to retrieve. One of: "bathymetry","infrastructure","land","land_cover",land_use","water"
    columns : list[str] | None, optional
        Specific columns to include in the result.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, optional
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing base data of the specified type within the bounding box
    """
    _check_base_type(base_type)
    complete_prefix = OVERTURE_BASE_PREFIX.format(type=base_type)
    return from_bbox(bbox,complete_prefix,OVERTURE_MAIN_PATH,OVERTURE_REGION,release,columns,filters)

def get_categories() -> DataFrame:
    df = read_csv(OVERTURE_CATEGORIES_URL, converters={" Overture Taxonomy": lambda x : x.strip().strip("[]").split(",")}, sep=';')
    
    # Create a dictionary to store unique IDs for each category
    category_ids = {}
    
    all_records = []
    
    for _, row in df.iterrows():
        category_hierarchy = row[" Overture Taxonomy"]
        
        # Only create a record for the last/deepest level in the hierarchy
        deepest_level = len(category_hierarchy)
        current_category = category_hierarchy[deepest_level - 1]
        
        for category in category_hierarchy:
            if category not in category_ids:
                u_id = str(uuid4()).replace('-', '')[:24]
                while u_id in category_ids.values():
                    u_id = str(uuid4()).replace('-', '')[:24]
                category_ids[category] = u_id
        
        # Build the full path string
        path_parts = []
        for category in category_hierarchy:
            formatted_part = ' '.join(word.capitalize() for word in category.split('_'))
            path_parts.append(formatted_part)
        
        full_path = ' > '.join(path_parts)
        
        record = {
            'category_id': category_ids[current_category],
            'category_level': deepest_level,
            'category_name': current_category,
            'category_label': full_path,
            'level1_category_id': None,
            'level1_category_name': None,
            'level2_category_id': None,
            'level2_category_name': None,
            'level3_category_id': None,
            'level3_category_name': None,
            'level4_category_id': None,
            'level4_category_name': None,
            'level5_category_id': None,
            'level5_category_name': None,
            'level6_category_id': None,
            'level6_category_name': None
        }
        
        # Fill in the level fields based on the hierarchy
        for j in range(min(deepest_level, 6)):
            level_name = category_hierarchy[j]
            level_id = category_ids[level_name]
            
            record[f'level{j+1}_category_id'] = level_id
            record[f'level{j+1}_category_name'] = level_name
        
        all_records.append(record)
    
    result_df = DataFrame(all_records)
    
    column_order = [
        'category_id', 'category_level', 'category_name', 'category_label',
        'level1_category_id', 'level1_category_name',
        'level2_category_id', 'level2_category_name',
        'level3_category_id', 'level3_category_name',
        'level4_category_id', 'level4_category_name',
        'level5_category_id', 'level5_category_name',
        'level6_category_id', 'level6_category_name'
    ]
    
    
    available_columns = [col for col in column_order if col in result_df.columns]
    
    result_df = result_df[available_columns]
    
    return result_df

def get_schema(dataset_name : str,
                connector : bool = False,
                building_part : bool = False,base_type : str = None,release: str = OVERTURE_LATEST_RELEASE) -> DataFrame:
    """
    Get DuckDB schema for the given dataset.

    Parameters
    ----------
        dataset_name : str
            Name of the dataset to get the schema of, must be one of \"buildings\",\"transportation\",\"base\",\"places\",\"addresses\"
        connector : bool, optional
            Whether to retrieve connector schema. Defaults to False.
        building_part : bool, optional 
            Whether to retrieve building_part schema. Defaults to False.
        base_type : str, optional 
            Which base type schema to retrive.
        release : str, optional 
            Release version to use, defaults to the latest version.
    Returns
    -------
    DataFrame
        DataFrame with columns showing DuckDB types
    """
    datasets = ["buildings","transportation","base","places","addresses"]
    with resources.files("pyplaces").joinpath("releases/overture/base_types.txt").open( "r",encoding="utf-8-sig") as f:
        base_types = [line.replace("type=", "").strip(" \n/") for line in f]
    if dataset_name not in datasets:
        raise KeyError(f"No dataset: {dataset_name} found")
    elif connector and dataset_name != "transportation":
        raise KeyError("Dataset must be \"transportation\" to get connector schema")
    elif building_part and dataset_name != "buildings":
        raise KeyError("Dataset must be \"buildings\" to get building_part schema")
    elif base_type and dataset_name != "base":
        raise KeyError(f"Dataset must be \"base\" to get {base_type} schema")
    elif base_type and base_type not in base_types:
        raise KeyError(f"No base type:{dataset_name} found")
    
    path = OVERTURE_MAIN_PATH.format(release=release)
    if dataset_name == "places":
        path = path + OVERTURE_PLACES_PREFIX
    elif dataset_name == "addresses":
        path = path + OVERTURE_ADDRESSES_PREFIX
    elif dataset_name == "buildings":
        if building_part:
            path = path + OVERTURE_BUILDINGS_PART_PREFIX
        else:
            path = path + OVERTURE_BUILDINGS_PREFIX
    elif dataset_name == "base":
        path = path + OVERTURE_BASE_PREFIX.format(type=base_type)
    elif dataset_name == "transportation":
        if connector:
            path = path + OVERTURE_TRANSPORTATION_CONNECTOR_PREFIX
        else:
            path = path + OVERTURE_TRANSPORTATION_SEGMENT_PREFIX
    schema = schema_from_dataset(path,OVERTURE_REGION)
    return schema

def find_categories(search: str, num_results: int = 5, exact_match: bool=False,verbose: bool=False,as_df: bool= False) -> Union[list[str],DataFrame]:
    """
    Finds Overture Places categories based on a user search.
    
    Parameters
    ----------
    search : str
        User search term for matching.
    num_results : int, optional. Defaults to 5
        Number of matched categories to retrieve.
    exact_match : bool, optional. Defaults to False.
        Whether to retrieve only exact matches from search.
    verbose : bool, optional. Defaults to False.
        Whether to show the additional information of the matches
    as_df : bool, optional. Defaults to False.
        Whether to retrieve the matches as a DataFrame with additional information.
    
    Returns
    -------
    Union[list[str],Dataframe]
        Matched Overture Category names as a list of strings or DataFrame.
    """
    finder = CategoryFinder()
    categories = get_categories()
    finder.load_data(categories)
    finder.process_data()
    matches = finder.suggest_categories(search,num_results,exact_match,verbose,as_df,hide_ids=True,list_return="name",show_name_and_id=False)
    return matches

#TODO no automation
def _check_base_type(base_type):
    """
    Validate if the provided base type is valid.
    
    Parameters
    ----------
    base_type : str
        The base type to validate
        
    Raises
    ------
    ValueError
        If the base type is not valid
    """   
    with resources.files("pyplaces").joinpath("releases/overture/base_types.txt").open( "r",encoding="utf-8-sig") as f:
        base_types = [line.replace("type=", "").strip(" \n/") for line in f]
    if base_type not in base_types:
        raise ValueError(f"Invalid base type:{base_type}")
    

    
__all__ = ["overture_addresses_from_address","overture_addresses_from_bbox","overture_addresses_from_place","overture_base_from_address", 
            "overture_base_from_bbox","overture_base_from_place","overture_buildings_from_address","overture_buildings_from_bbox",
            "overture_buildings_from_place","overture_places_from_address","overture_places_from_bbox","overture_places_from_place",
            "overture_transportation_from_address","overture_transportation_from_bbox","overture_transportation_from_place","get_schema","find_categories","get_categories"]

wrap_functions_with_release(__name__, _check_or_get_release,__all__)



"""Functions to fetch geoparquet data from OSM Layercake datasets"""
from geopandas import GeoDataFrame
from pandas import DataFrame, json_normalize, concat
from ._io_utils import from_address, from_bbox, from_place, schema_from_dataset

# OSM Layercake dataset URLs
LAYERCAKE_BASE_URL = "https://data.openstreetmap.us/layercake/"
# OSM_DATASETS = {
#     "buildings": f"{LAYERCAKE_BASE_URL}buildings.parquet",
#     "highways": f"{LAYERCAKE_BASE_URL}highways.parquet"
# }

def layercake_buildings_from_address(address: str | tuple[float, float],
                                columns: list[str] | None = None,
                                filters: str | None = None,
                                distance: float = 500,
                                unit: str = "m") -> GeoDataFrame:
    """
    Retrieves OSM Layercake buildings data in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float, float]
        The address or (longitude, latitude) tuple to search for nearby buildings.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
    distance : float, default 500
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, default "m"
        Unit of measurement for the distance. Defaults to "m" (meters). 
        One of: "m", "km", "in", "ft", "yd", "mi"
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM buildings data within the specified bounding box.
    """
    gdf = from_address(address, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters, distance, unit,"buildings")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf


def layercake_buildings_from_place(address: str,
                            columns: list[str] | None = None,
                            filters: str | None = None) -> GeoDataFrame:
    """
    Retrieves OSM Layercake buildings data for a specific place identified by address or name.
    
    Parameters
    ----------
    address : str
        The address or identifier of the place to retrieve data for.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM buildings data within specified place.
    """
    gdf = from_place(address, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters,"buildings")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf


def layercake_buildings_from_bbox(bbox: tuple[float, float, float, float],
                            columns: list[str] | None = None,
                            filters: str | None = None) -> GeoDataFrame:
    """
    Retrieves OSM Layercake buildings data within a specified bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        Bounding box coordinates in the format (min_x, min_y, max_x, max_y).
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM buildings data within the specified bounding box.
    """
    gdf = from_bbox(bbox, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters,"buildings")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf


def layercake_highways_from_address(address: str | tuple[float, float],
                            columns: list[str] | None = None,
                            filters: str | None = None,
                            distance: float = 500,
                            unit: str = "m") -> GeoDataFrame:
    """
    Retrieves OSM Layercake highways data in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float, float]
        The address or (longitude, latitude) tuple to search for nearby highways.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
    distance : float, default 500
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, default "m"
        Unit of measurement for the distance. Defaults to "m" (meters). 
        One of: "m", "km", "in", "ft", "yd", "mi"
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM highways data within the specified bounding box.
    """
    gdf = from_address(address, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters, distance, unit,"highways")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf
    


def layercake_highways_from_place(address: str,
                            columns: list[str] | None = None,
                            filters: str | None = None) -> GeoDataFrame:
    """
    Retrieves OSM Layercake highways data for a specific place identified by address or name.
    
    Parameters
    ----------
    address : str
        The address or identifier of the place to retrieve data for.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM highways data within specified place.
    """
    gdf = from_place(address, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters,"highways")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf


def layercake_highways_from_bbox(bbox: tuple[float, float, float, float],
                            columns: list[str] | None = None,
                            filters: str | None = None) -> GeoDataFrame:
    """
    Retrieves OSM Layercake highways data within a specified bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        Bounding box coordinates in the format (min_x, min_y, max_x, max_y).
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression for filtering results.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing OSM highways data within the specified bounding box.
    """
    gdf = from_bbox(bbox, "", LAYERCAKE_BASE_URL, None, 
                    None, columns, filters,"highways")
    tags_df = json_normalize(gdf['tags'])
    gdf = concat([
        gdf.iloc[:, :2],           # First 2 columns
        tags_df,                    # Unpacked tags
        gdf.iloc[:, 2:].drop(columns=['tags'])  # Rest of columns (excluding tags)
    ], axis=1)
    return gdf


def get_schema(dataset: str = "buildings") -> DataFrame:
    """
    Get DuckDB schema for the specified OSM Layercake dataset.

    Parameters
    ----------
    dataset : str, default "buildings"
        The dataset to get schema for. One of: "buildings", "highways"
        
    Returns
    -------
    DataFrame
        DataFrame with columns showing DuckDB types.
        
    Raises
    ------
    ValueError
        If the specified dataset is not valid.
    """
    if dataset not in ["buildings","highways"]:
        raise ValueError(f"Invalid dataset: {dataset}. Must be one of {["buildings","highways"]}")
    
    path = LAYERCAKE_BASE_URL + dataset
    schema = schema_from_dataset(path, None)
    return schema


__all__ = [
    "layercake_buildings_from_address",
    "layercake_buildings_from_bbox", 
    "layercake_buildings_from_place",
    "layercake_highways_from_address",
    "layercake_highways_from_bbox",
    "layercake_highways_from_place",
    "get_schema"
]


"""Functions to fetch geoparquet data from Foursquare Open Places on AWS"""
from importlib import resources
import requests
from io import StringIO
from typing import Union
from geopandas import GeoDataFrame
from pandas import DataFrame
from ._utils import wrap_functions_with_release
from ._io_utils import from_address, from_bbox, from_place, read_parquet_duckdb, schema_from_dataset
from ._category_finder import CategoryFinder

FSQ_VERSIONS_URL = "https://github.com/tbanken/pyplaces/blob/main/src/pyplaces/releases/foursquare/releases.txt"

def _check_or_get_release(release: str = None, url: str = FSQ_VERSIONS_URL ,latest: bool = False):
    """
    Validates if the specified Foursquare data release version exists.
    
    Parameters
    ----------
    release : str
        The release version to validate.
    
    url : str
        GitHub url to file version.
        
    Raises
    ------
    ValueError
        If the specified release version does not exist in the available releases.
    """
    raw_url = url.replace('github.com', 'raw.githubusercontent.com').replace('/blob/', '/')
    
    try:
        response = requests.get(raw_url,timeout=10)
        response.raise_for_status()
        file = StringIO(response.text)
        
        folders = [line.replace("dt=", "").strip(" \n/") for line in file]
    except requests.RequestException as e:
        print(f"Error fetching file: {e}, using local version check")
        if latest:
            print("Using local version file. Current release may not be up to date.")
        with resources.files("pyplaces").joinpath("releases/foursquare/releases.txt").open("r", encoding="utf-8-sig") as file:
            folders = [line.replace("dt=", "").strip(" \n/") for line in file]
    folders.remove("vector-tiles")
    if latest:
        return folders[-1]
    if release not in folders:
        raise ValueError(f"Invalid release: {release}")
    

FSQ_LATEST_RELEASE = _check_or_get_release(latest=True)

FSQ_MAIN_PATH = 's3://fsq-os-places-us-east-1/release/dt={release}/'
FSQ_BUCKET = 'fsq-os-places-us-east-1'
FSQ_REGION = 'us-east-1'
FSQ_PLACES_PREFIX = "places/parquet/"
FSQ_CATEGORIES_PREFIX = "categories/parquet/"
# FSQ_LATEST_RELEASE = "2025-08-07"

# FSQ_FUSED_MAIN_PATH = 's3://us-west-2.opendata.source.coop/fused/fsq-os-places/{release}/'
# FSQ_FUSED_BUCKET = 's-west-2.opendata.source.coop'
# FSQ_FUSED_REGION = 'us-west-2'
# FSQ_FUSED_LATEST_RELEASE = "2025-04-08"
# FSQ_FUSED_PLACES_PREFIX = "places/"

def foursquare_places_from_address(address: str | tuple[float,float],
                                    columns: list[str] | None = None,
                                    filters: str | None = None,
                                    distance: float = 500,
                                    unit: str = "m",
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data in a bounding box around a specified address.
    
    Parameters
    ----------
    address : str | tuple[float,float]
        The addres or (longitude, latitude) tuple to search for nearby places.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression
    distance : float, default 500
        Radius of the bounding box around the address. Defaults to 500 meters.
    unit : str, default "m"
        Unit of measurement for the distance. Defaults to "m" (meters). One of: "m","km","in","ft","yd","mi"
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified bounding box of the address.
    """
    return from_address(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters, distance, unit)


def foursquare_places_from_place(address: str,
                                    columns: list[str] | None = None,
                                    filters: str = None,
                                    release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data for a specific place identified by its address or place name.
    
    Parameters
    ----------
    address : str
        The address or identifier of the place to retrieve data for.
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within specified place.
    """
    return from_place(address, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def foursquare_places_from_bbox(bbox: tuple[float, float, float, float],
                                columns: list[str] | None = None,
                                filters: str | None = None,
                                release: str = FSQ_LATEST_RELEASE) -> GeoDataFrame:
    """
    Retrieves Foursquare places data within a specified bounding box.
    
    Parameters
    ----------
    bbox : tuple[float, float, float, float]
        Bounding box coordinates in the format (min_x, min_y, max_x, max_y).
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset. 
    filters : str | None, optional
        DuckDB SQL expression
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing Foursquare places data within the specified bounding box.
    """
    return from_bbox(bbox, FSQ_PLACES_PREFIX, FSQ_MAIN_PATH, FSQ_REGION, release, columns, filters)


def get_categories(columns: list[str] | None = None,
                    filters: str | None = None,
                    release: str = FSQ_LATEST_RELEASE) -> DataFrame:
    """
    Retrieves Foursquare place categories data.
    
    Parameters
    ----------
    columns : list[str] | None, optional
        Specific columns to retrieve from the dataset.
    filters : str | None, optional
        DuckDB SQL expression
    release : str, default FSQ_LATEST_RELEASE
        Dataset release version to use. Defaults to the latest version.
        
    Returns
    -------
    DataFrame
        A DataFrame containing Foursquare place categories data.
        
    Notes
    -----
    This function constructs the data path using Foursquare-specific constants and
    retrieves the categories data using read_parquet_duckdb.
    """
    path = FSQ_MAIN_PATH.format(release=release) + FSQ_CATEGORIES_PREFIX
    return read_parquet_duckdb(path, FSQ_REGION, columns, filters)

def find_categories(search: str, num_results: int = 5, exact_match: bool=False,verbose: bool=False,as_df: bool= False) -> Union[list[str],DataFrame]:
    """
    Finds Foursquare Open Places categories based on a user search.
    
    Parameters
    ----------
    search : str
        User search term for matching.
    num_results : int, optional. Defaults to 5
        Number of matched categories to retrieve.
    exact_match : bool, optional. Defaults to False.
        Whether to retrieve only exact matches from search.
    verbose : bool, optional. Defaults to False.
        Whether to show the additional information of the matches
    as_df : bool, optional. Defaults to False.
        Whether to retrieve the matches as a DataFrame with additional information.
    
    Returns
    -------
    Union[list[str],Dataframe]
        Matched Foursquare Category IDs as a list of strings or DataFrame.
    """
    finder = CategoryFinder()
    categories = get_categories()
    finder.load_data(categories)
    finder.process_data()
    matches = finder.suggest_categories(search,num_results,exact_match,verbose,as_df,hide_ids=False,list_return="",show_name_and_id=True)
    return matches

def get_schema(categories=False,release:str=FSQ_LATEST_RELEASE) -> DataFrame:
    """
    Get DuckDB schema for the given dataset. Set categories to True if you want to get the categories instead of the places schema.

    Parameters
    ----------
        connector : bool, optional
            Whether to get categories schema, by default False.
        release : str, optional 
            Release version to use, defaults to the latest version.
    Returns
    -------
    DataFrame
        DataFrame with columns showing DuckDB types
    """
    path = FSQ_MAIN_PATH.format(release=release) 
    if categories:
        path = path + FSQ_CATEGORIES_PREFIX
    else:
        path = path + FSQ_PLACES_PREFIX
    schema = schema_from_dataset(path,FSQ_REGION)
    return schema

__all__ = ["foursquare_places_from_address", "foursquare_places_from_bbox", "foursquare_places_from_place", "get_categories","get_schema","find_categories"]

wrap_functions_with_release(__name__, _check_or_get_release,__all__)



"""I/O utility functions for reading data."""

from __future__ import annotations
import json
from geopandas import GeoDataFrame
import duckdb
import geopandas as gpd


from ._geo_utils import geocode_place_to_bbox, geocode_point_to_bbox

def schema_from_dataset(path,region,file_name: str | None = None):
    """
    Get schema from parquet dataset.
    
    Parameters:
    -----------
    s3_path : str 
        S3 path to directory containing parquet dataset
    region : str
        AWS region
    Returns:
    --------
    Schema
        DuckDB schema from given dataset.
    """
    conn = duckdb.connect()
    if region:
        conn.execute(f"SET s3_region='{region}';")
        conn.execute("SET s3_use_ssl=true;")
    if file_name:
        path = path + f"{file_name}.parquet"
    else:
        path = path + '*.parquet'
    
    schema_query = f"""
        DESCRIBE SELECT *
        FROM '{path}';
        """
    df = conn.execute(schema_query).df().iloc[:,0:2]
    conn.close()
    return df

def read_geoparquet_duckdb(path: str, region: str, bbox: tuple[float,float,float,float], 
                        columns: list[str] | None = None, 
                        filters: str | None = None,
                        file_name: str | None = None) -> GeoDataFrame:
    """
    Read geospatial data from a parquet file on S3 with filtering by bbox.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    columns : list, optional
        Columns to select
    filters : str, optional
        DuckDB SQL expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    conn = duckdb.connect()
    
    # Install and load spatial extension for geometry handling
    conn.execute("INSTALL spatial;")
    conn.execute("LOAD spatial;")
    conn.execute("INSTALL httpfs;")
    conn.execute("LOAD httpfs;")
    if region:
        conn.execute(f"SET s3_region='{region}';")
        conn.execute("SET s3_use_ssl=true;")
    if file_name:
        path = path + f"{file_name}.parquet"
    else:
        path = path + '*.parquet'
    
    xmin, ymin, xmax, ymax = bbox
    
    metadata_query = f"""
                    SELECT decode(key) as 'key', decode(value) as 'value'
                    FROM parquet_kv_metadata('{path}')
                    WHERE key = 'geo' 
                    LIMIT 1;
                    """
    metadata_df = conn.execute(metadata_query).df()
    geom_col_name = json.loads(metadata_df['value'].iloc[0])['primary_column']
    
    # Select all if no selections, otherwise join selections together, exclude geometry column and cast it as text column
    column_select = ("*" if columns is None else ", ".join(columns)) + f" EXCLUDE ({geom_col_name}), ST_AsText({geom_col_name}) as 'geometry'"
    
    # Build spatial filter using bbox columns
    spatial_filter = f"""
        bbox.xmin < {xmax} AND
        bbox.xmax > {xmin} AND
        bbox.ymin < {ymax} AND
        bbox.ymax > {ymin}
    """
    
    # Construct the main query
    query = f"""
        SELECT {column_select}
        FROM read_parquet('{path}')
        WHERE {spatial_filter}
    """
    if filters:
        query = query + f" AND {filters}"
    try:
        df = conn.execute(query).df()
    except Exception as e:
        raise e
        # raise ValueError("Filter or Column error. Either the filter is not formatted properly, or the wrong column/value name is used.") from e
        
    
    conn.close()
    
    gdf = gpd.GeoDataFrame(df,crs="EPSG:4326",geometry=gpd.GeoSeries.from_wkt(df["geometry"]))  
    
    return gdf

def read_parquet_duckdb(path: str, region: str, 
                    columns: list[str] | None = None, 
                    filters: str | None = None,
                    file_name: str | None = None) -> GeoDataFrame:
    """
    Read tabular data from a parquet file on S3.
    
    Parameters:
    -----------
    path : str
        S3 path to the parquet file
    region : str
        AWS region
    columns : list, optional
        Columns to select
    filters : str, optional
        DuckDB SQL expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered dataframe
    """
    conn = duckdb.connect()
    
    # Install and load spatial extension for geometry handling
    conn.execute("INSTALL httpfs;")
    conn.execute("LOAD httpfs;")
    if region:
        conn.execute(f"SET s3_region='{region}';")
        conn.execute("SET s3_use_ssl=true;")
    if file_name:
        path = path + f"{file_name}.parquet"
    else:
        path = path + '*.parquet'
    
    # Select all if no selections, otherwise join selections together, exclude geometry column and cast it as text column
    column_select = "*" if columns is None else ", ".join(columns)
    
    # Construct the main query
    query = f"""
        SELECT {column_select}
        FROM read_parquet('{path}')
    """
    if filters:
        query = query + f" WHERE {filters}"
        
    try:
        df = conn.execute(query).df()
    except Exception as e:
        # print(e)
        raise ValueError("Filter or Column error. Either the filter is not formatted properly, or the wrong column/value name is used.") from e
        
    conn.close()
    
    return df

def _get_gdf_from_bbox(release:str, bbox:tuple[float,float,float,float], columns:list[str], filters: str, prefix: str, path: str, region: str, file_name: str | None = None):
    """Helper function to get a geodataframe from a bounding box."""
    main_path = path.format(release=release) + prefix
    gdf = read_geoparquet_duckdb(main_path, region, bbox, columns=columns, filters=filters,file_name=file_name)
    return gdf

def from_address(address: str | tuple[float,float], prefix: str, main_path: str, region: str,
            release: str, columns: list[str]| None = None, filters: str | None = None,
            distance: float = 500, unit: str = "m",  file_name: str | None = None) -> GeoDataFrame:
    """
    Wrapper to geocode an address and fetch the geoparquet data within the address's area.
    
    Parameters:
    -----------
    address : str or tuple
        Address string or (longitude, latitude) tuple
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : str, optional
        DuckDB SQL expression
    distance : float, default 500
        Buffer distance
    unit : str, default 'm'
        Unit of distance
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    bbox = geocode_point_to_bbox(address, distance, unit)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters,file_name)
    return gdf
    
def from_place(address: str, prefix: str, main_path: str, region: str, release: str,
            columns: list[str]| None=None, filters: str| None=None,  file_name: str | None = None) -> GeoDataFrame:
    """
    Wrapper to geocode a place and fetch the geoparquet data within the place.
    
    Parameters:
    -----------
    address : str
        Place name or address
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : str, optional
        DuckDB SQL expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    geometry, bbox = geocode_place_to_bbox(address)
    gdf = from_bbox(bbox,prefix,main_path,region,release,columns,filters,file_name)
    filtered_gdf = gdf[gdf.within(geometry)]
    return filtered_gdf

def from_bbox(bbox: tuple[float,float,float,float], prefix: str, main_path: str, region: str, 
            release: str, columns: list[str]| None=None, filters: str | None=None,  file_name: str | None = None) -> GeoDataFrame:
    """
    Wrapper to fetch the geoparquet data within the bounding box.
    
    Parameters:
    -----------
    bbox : tuple
        Bounding box as (minx, miny, maxx, maxy)
    prefix : str
        Path prefix for the data
    main_path : str
        Base path template with {release} placeholder
    region : str
        AWS region
    release : str
        Release version
    columns : list, optional
        Columns to select
    filters : str, optional
        DuckDB SQL expression
        
    Returns:
    --------
    GeoDataFrame
        Filtered geodataframe
    """
    gdf = _get_gdf_from_bbox(release, bbox, columns, filters, prefix, main_path, region,file_name)
    return gdf


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
# import sys
# from pathlib import Path
# sys.path.insert(0, str(Path('../..', 'src').resolve()))


project = 'pyplaces'
copyright = '2025, Theodore Banken'
author = 'Theodore Banken'
release = '0.4.5'
#TODO dynamically get release
# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc','sphinx_autodoc_typehints','nbsphinx']

templates_path = ['_templates']
exclude_patterns = []

# html_theme_options = {
#     "home_page_in_toc": True,
#     "collapse_navbar": True,
# }
nbsphinx_execute = 'never'

suppress_warnings = [
    'nbsphinx',
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']



import re
import math
from typing import List, Dict, Union
from pandas import notna, DataFrame
from sentence_transformers import SentenceTransformer

class CategoryFinder:
    """
    A class to find relevant categories based on user queries.
    Uses semantic matching for more accurate results.
    """
    
    def __init__(self, batch_size: int = 1000):
        self.categories_df = None
        self.batch_size = batch_size
        self.processed = False
        
        # Initialize semantic model
        
        self.semantic_model = SentenceTransformer('all-mpnet-base-v2')
        self.semantic_vectors = []
    
    
    def load_data(self, data: DataFrame):
        """
        Load category data, either from a DataFrame or from a file path.
        
        Parameters:
        -----------
        data: 
            DataFrame containing category data or path to CSV file
        Returns:
        --------
        self
        """
        self.categories_df = data
            
        self.categories_df.columns = [col.strip() for col in self.categories_df.columns]
        
        # Create cleaned category names
        self.categories_df['category_name_clean'] = self.categories_df['category_name'].fillna('').apply(
            lambda x: re.sub(r'\s+', ' ', x.lower()).strip()
        )
        
        return self
    
    
    def process_data(self):
        """
        Process the category data to build the semantic embeddings.
        
        Returns:
        --------
        self
        """
        if self.categories_df is None:
            raise ValueError("Data must be loaded before processing. Use load_data() first.")
        
        # Process semantic embeddings
        texts_to_embed = self.categories_df['category_name_clean'].tolist()
        
        self.semantic_vectors = []
        for start_idx in range(0, len(texts_to_embed), self.batch_size):
            end_idx = min(start_idx + self.batch_size, len(texts_to_embed))
            batch_texts = texts_to_embed[start_idx:end_idx]
            batch_vectors = self.semantic_model.encode(batch_texts)
            self.semantic_vectors.extend(batch_vectors)
        
        self._create_lookup_tables()
        
        self.processed = True
        return self
    
    def _cosine_similarity(self, vec1: List[float], vec2: List[float]) -> float:
        dot_product = sum(a * b for a, b in zip(vec1, vec2))
        
        # Calculate the magnitudes
        mag1 = math.sqrt(sum(a * a for a in vec1))
        mag2 = math.sqrt(sum(b * b for b in vec2))
        
        if mag1 == 0 or mag2 == 0:
            return 0
        
        # Return cosine similarity
        return dot_product / (mag1 * mag2)
        
    def _create_lookup_tables(self):
        self.category_name_lookup = {
            name.lower(): idx for idx, name in 
            enumerate(self.categories_df['category_name_clean'])
        }
        
        self.parent_child_map = {}
        for i in range(1, 6): 
            parent_id_col = f'level{i}_category_id'
            if parent_id_col in self.categories_df.columns:
                for _, row in self.categories_df.iterrows():
                    if notna(row.get(parent_id_col)):
                        parent_id = row[parent_id_col]
                        if parent_id not in self.parent_child_map:
                            self.parent_child_map[parent_id] = []
                        if notna(row.get('category_id')):
                            self.parent_child_map[parent_id].append(row['category_id'])
    
    def _find_categories(self, query: str, top_n: int = 5, threshold: float = 0.3, 
                        exact_match: bool = False) -> List[Dict]:
        if not self.processed:
            raise ValueError("Data must be processed before searching. Use process_data() first.")
        
        query = query.lower().strip()
        
        exact_matches = []
        normalized_query = query
        
        
        if normalized_query in self.category_name_lookup:
            idx = self.category_name_lookup[normalized_query]
            category = self.categories_df.iloc[idx]
            exact_matches.append({
                'category_id': category.get('category_id'),
                'category_name': category.get('category_name'),
                'category_label': category.get('category_label', ''),
                'similarity_score': 1.0,  # Perfect match
                'category_level': int(category.get('category_level', 0)),
                'match_type': 'exact'
            })
        
        
        if exact_matches and exact_match:
            return exact_matches
            
        
        results = exact_matches.copy()
        added_indices = {self.category_name_lookup.get(m['category_name'].lower(), -1) for m in exact_matches}
        
        # Create semantic vector for query
        query_semantic = self.semantic_model.encode([query])[0]
        
        
        semantic_similarities = [self._cosine_similarity(query_semantic, vec) for vec in self.semantic_vectors]
        
        similarity_pairs = [(idx, score) for idx, score in enumerate(semantic_similarities)]
        similarity_pairs.sort(key=lambda x: x[1], reverse=True)
        
        for idx, score in similarity_pairs:
            if len(results) >= top_n:
                break
                
            if score >= threshold and idx not in added_indices:
                added_indices.add(idx)
                category = self.categories_df.iloc[idx]
                results.append({
                    'category_id': category.get('category_id'),
                    'category_name': category.get('category_name'),
                    'category_label': category.get('category_label', ''),
                    'similarity_score': float(score),
                    'category_level': int(category.get('category_level', 0)),
                    'match_type': 'semantic'
                })
                
        return sorted(results, key=lambda x: x['similarity_score'], reverse=True)
    
    def suggest_categories(self, query: str, top_n, exact_match,verbose,as_df,hide_ids,list_return,show_name_and_id=False) -> Union[list[str], DataFrame]:
        """
        Suggest categories using exact matching and semantic matching.
        
        Parameters
        ----------
        query : str
            Category search query from user
        top_n : int
            Number of matches to return
        exact_match : bool, optional
            Whether to return exact matches only. Defaults to False
        verbose : bool, optional
            Whether to return matches with additional information. Defaults to False
        as_df : bool, optional
            Whether to return matches as a dataframe with additional information
            
        Returns
        -------
        Union[list[str], DataFrame]
            List or DataFrame of matches
        """
        
        matches = self._find_categories(query, top_n=top_n, exact_match=exact_match)
        
        # matches = self._remove_redundant_children(matches)
        df = DataFrame(matches)
        if hide_ids:
            df = df[df.columns.drop(list(df.filter(regex='category_id')))]
        
        if show_name_and_id:
            if verbose:
                print(df)
            else:
                print(df[["category_id","category_name"]])    
        if as_df:
            return df
        if list_return == "name":
            return [item['category_name'] for item in matches]
        else:
            return [item['category_id'] for item in matches]
    
    def _remove_redundant_children(self, categories: List[Dict]) -> List[Dict]:
        if not categories or len(categories) <= 1:
            return categories
            
        category_ids = {cat['category_id'] for cat in categories}
        
        children_to_remove = set()
        
        for category in categories:
            level = category.get('category_level', 0)
            category_id = category.get('category_id')
            
            for i in range(1, level):
                parent_level_col = f'level{i}_category_id'
                if parent_level_col in self.categories_df.columns:
                    cat_row = self.categories_df[self.categories_df['category_id'] == category_id]
                    if not cat_row.empty and parent_level_col in cat_row.columns:
                        parent_id = cat_row.iloc[0][parent_level_col]
                        # If the parent is in our results, mark this child for removal
                        if parent_id in category_ids:
                            children_to_remove.add(category_id)
                            break
        return [cat for cat in categories if cat['category_id'] not in children_to_remove]